# MIT License

# Copyright (c) 2016 Morgan McDermott & John Carlyle

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class RedLeaderError(Exception):
    """
    Base exception for all redleader exceptions
    """

    message = 'An unknown error has occured'

    def __init__(self, **kwargs):
        msg = self.message.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class MissingDependencyError(RedLeaderError):
    message = 'Resource {source_resource} missing dependency {missing_resource}. Ensure {source_resource} has been added to the cluster.'

class OfflineContextError(RedLeaderError):
    message = 'Attempting to use offline context like an online context.'

class OfflineContextError(RedLeaderError):
    message = 'OfflineContext cannot perform action {action}'



from .resource import Resource
from .ec2 import EC2InstanceResource
from .iam import IAMRoleResource, IAMInstanceProfileResource
import pkg_resources

class CodeDeployEC2InstanceResource(EC2InstanceResource):
    """
    Creates an EC2 server resource with the code deploy agent running.
    Takes the same keyword parameters as EC2InstanceResource.

    Automatically generates and depends upon a CodeDeployServiceRole.
    """

    def _get_init_script(self, region_name):
        resource_package = __name__
        resource_path = '/'.join(('scripts', 'code_deploy_init.sh'))
        s = pkg_resources.resource_string(resource_package, resource_path).decode('utf-8')
        s = s.replace("{region_name}", region_name)
        return s

    def generate_sub_resources(self):
        res = super().generate_sub_resources()
        if self._service_role is None:
            self._service_role = CodeDeployServiceRoleResource(self._context)
            self.add_dependency(self._service_role)

        res.append(self._service_role)
        res.append(self._instance_profile)
        self.add_dependency(self._instance_profile)
        return res

    def __init__(self, context, deployment_group, **kwargs):
        self._deployment_group = deployment_group
        self._service_role = None
        kwargs = self.configure_kwargs(context, kwargs)

        super().__init__(context, **kwargs)
        self.add_dependency(self._deployment_group)

    def configure_kwargs(self, context, kwargs):
        """
        Configure tags, init_script, and roles for code deploy
        """
        region_name = "us-west-1"
        if "region_name" in kwargs:
            region_name = kwargs[region_name]

        kwargs['tags'] = [self._deployment_group.get_ec2_tag()]
        kwargs['init_script'] = self._get_init_script(region_name)

        # Setup instance profile
        permissions = kwargs.get('permissions', [])
        roles = kwargs.get('roles', [])
        services = kwargs.get('services', []) +\
                   ['ec2.amazonaws.com', 'codedeploy.amazonaws.com']
        #roles.append(CodeDeployServiceRoleResource.global_service_role(context))

        self._instance_profile = IAMInstanceProfileResource(
            context,
            permissions=permissions,
            roles=roles,
            services=services
        )
        kwargs['instance_profile'] = self._instance_profile
        return kwargs

class CodeDeployServiceRoleResource(IAMRoleResource):
    def __init__(self, context, permissions=[], policies=[], policy_arns=[]):
        services = ['ec2.amazonaws.com', 'codedeploy.amazonaws.com']
        super().__init__(context,
                         permissions=permissions,
                         services=services,
                         policies=policies,
                         policy_arns = policy_arns +\
                           ["arn:aws:iam::aws:policy/service-role/AWSCodeDeployRole"]
        )

class CodeDeployDeploymentGroupResource(Resource):
    def __init__(self, context, application_name,
                 deployment_group_name=None, cf_params={}):
        super().__init__(context, cf_params)
        self._application_name = application_name
        self._application = None
        self._service_role = CodeDeployServiceRoleResource(self._context)
        self.add_dependency(self._service_role)
        self._ec2_tag_filters = [
            {"Key": "redleaderDeploymentGroup",
             "Type": "KEY_AND_VALUE",
             "Value": self._id_placeholder()
            }
        ]

        if deployment_group_name is None:
            self._deployment_group_name = self._id_placeholder()
        else:
            self._deployment_group_name = deployment_group_name

    def get_ec2_tag(self):
        return {"Key": "redleaderDeploymentGroup", "Value": self.get_id()}

    def generate_sub_resources(self):
        res = super().generate_sub_resources()
        res.append(self._service_role)

        if self._application is None:
            self._application = CodeDeployApplicationResource(
                self._context, self._application_name)
            self.add_dependency(self._application)
        res.append(self._application)
        return res

    def _cloud_formation_template(self):
        return {
          "Type" : "AWS::CodeDeploy::DeploymentGroup",
          "Properties" : {
              "ApplicationName" : self._application_name,
              #"AutoScalingGroups" : [],
              #"Deployment" : Deployment,
              #"DeploymentConfigName" : String,
              "DeploymentGroupName": self._deployment_group_name,
              "Ec2TagFilters" : self._ec2_tag_filters,
              "ServiceRoleArn" : Resource.cf_attr(self._service_role, "Arn")
          }
        }

class CodeDeployApplicationResource(Resource):
    def __init__(self, context, application_name, cf_params={}):
        super().__init__(context, cf_params)
        self._application_name = application_name

    def _cloud_formation_template(self):
        return {
            "Type" : "AWS::CodeDeploy::Application",
            "Properties" : {
                "ApplicationName": self._application_name,

            }
        }



from redleader.resources import Resource
from .iam import IAMInstanceProfileResource
import random
import botocore.exceptions
import redleader.exceptions as exceptions

class EC2InstanceResource(Resource):
    """
    Creates an EC2 server resource.

    permissions (optional):
    If a list of permissions is provided, then an instance profile
    will be created with appropriate permissions.

    security_key_name (optional):
    A security key will automatically be created and stored if none is supplied

    security_group (optional): An EC2SecurityGroupResource
    instance_profile (optional): An IAMInstanceProfileResource
    storage_size (default=20): Size of default block storage
    ami (default=ami-165a0876): An AMI string
    instance_type (default=t2.micro): EC2 instance type
    init_script (optional): Script to be executed on EC2 server start
    """

    DEFAULTS = {
        "permissions": [],
        "security_group": None,
        "security_key_name": None,
        "instance_profile": None,
        "storage_size": 20,
        "ami": "ami-165a0876",
        "instance_type": "t2.micro",
        "region_name": "us-west-1",
        "init_script": "",
        "tags": [],
        "cf_params": {}
    }
    def __init__(self, context, **kwargs):
        for key, value in self.DEFAULTS.items():
            if key in kwargs:
                setattr(self, "_" + key, kwargs[key])
            else:
                setattr(self, "_" + key, value)

        super().__init__(context, self._cf_params)
        self._generated_profile = None

        for permission in self._permissions:
            self.add_dependency(permission.resource)

        if self._security_group is not None:
            self.add_dependency(self._security_group)

        if self._security_key_name is None:
            try:
                ec2 = self._context.get_client('ec2')
            except exceptions.OfflineContextError:
                return
            try:
                response = ec2.create_key_pair(KeyName='redleader_keypair')
                with open("./redleader_keypair.pem", "w") as f:
                    f.write(response['KeyMaterial'])
            except botocore.exceptions.ClientError:
                # Key already exists. We're good to go.
                pass
            self._security_key_name="redleader_keypair"

    def generate_sub_resources(self):
        generated = []

        if self._security_group is None:
            self._security_group = EC2SecurityGroupResource(
                self._context,
                ingressConfigs=[{
                    "IpProtocol" : "tcp",
                    "FromPort" : "22",
                    "ToPort" : "22",
                    "CidrIp" : "0.0.0.0/0"
                }]
            )
            generated.append(self._security_group)
            self.add_dependency(self._security_group)

        if self._instance_profile is not None:
            if self._generated_profile is None:
                self._generated_profile = IAMInstanceProfileResource(
                    self._context, self._permissions, roles=[])
                self.add_dependency(self._generated_profile)
            generated.append(self._generated_profile)

        return generated

    def _cloud_formation_block_device_mappings(self):
        return [{
            "DeviceName" : "/dev/sdm",
            "Ebs" : {
                "VolumeType" : "io1",
                "Iops" : "200",
                "DeleteOnTermination" : "true",
                "VolumeSize" : self._storage_size
            }
        }]

    def _cloud_formation_template(self):
        return {
            "Type" : "AWS::EC2::Instance",
            "Properties" : {
                "ImageId" : self._ami,
                "InstanceType": self._instance_type,
                "IamInstanceProfile": Resource.cf_ref(self._instance_profile),
                "KeyName" : self._security_key_name,
                "SecurityGroupIds": [Resource.cf_ref(self._security_group)],
                "BlockDeviceMappings" : self._cloud_formation_block_device_mappings(),
                "UserData": {"Fn::Base64": self._init_script},
                "Tags": self._tags
            }
        }
        raise NotImplementedError

class EC2SecurityGroupResource(Resource):
    """
    Expects ingress/egress objects in the format:
    {
      "IpProtocol" : "tcp",
      "FromPort" : "80",
      "ToPort" : "80",
      "CidrIp" : "0.0.0.0/0"
    }
    """
    def __init__(self, context, ingressConfigs=[], egressConfigs=[], cf_params={}):
        super().__init__(context, cf_params)
        self._ingressConfigs = ingressConfigs
        self._egressConfigs = egressConfigs

    def _cloud_formation_template(self):
        return {
            "Type" : "AWS::EC2::SecurityGroup",
            "Properties" : {
                #"VpcId" : {"Ref" : "myVPC"},
                "GroupDescription": "RedLeader security group %s" % self._id_placeholder(),
                "SecurityGroupIngress": self._ingressConfigs,
                "SecurityGroupEgress": self._egressConfigs
            }
        }



from redleader.resources import Resource

class DynamoDBTableResource(Resource):
    """
    Resource modeling a DynamoDB Table.

    If attribute_definitions and key_schema are not provided,
    the resource will not generate a cloud formation template
    and will be useful only for generating IAM policies/roles/profiles.

    key_schema example:
        OrderedDict([
            ('your_primary_key', 'HASH'),
            ('your_range_key', 'RANGE')
         ])
    attribute_definitions example:
         {
            'your_primary_key': 'S',
            'your_range_key': 'N',
            'some_other_column': 'S',
         }
    """
    def __init__(self,
                 context,
                 table_name,
                 attribute_definitions=None,
                 key_schema=None,
                 read_units=1,
                 write_units=1,
                 cf_params={}
    ):
        super(DynamoDBTableResource, self).__init__(context, cf_params)
        self._attribute_definitions = attribute_definitions
        self._key_schema = key_schema
        self._read_units = read_units
        self._write_units = write_units
        self._table_name = table_name

    def is_static(self):
        return True

    def get_id(self):
        return "DynamoDBTable%s" % self._table_name.replace("-", "").replace("_", "")

    def _iam_service_policy(self):
        return {"name": "dynamodb",
                "params": {"table_name": self._table_name}}

    def _cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        if self._key_schema is None or self._attribute_definitions is None:
            return None

        key_schema = []
        attribute_definitions = []
        for key in self._key_schema:
            key_schema.append({'AttributeName': key,
                              'KeyType': self._key_schema[key]})
        for key in self._attribute_definitions:
            if key in self._key_schema:
                attribute_definitions.append({'AttributeName': key,
                                              'AttributeType': self._attribute_definitions[key]})

        return {
            "Type" : "AWS::DynamoDB::Table",
            "Properties" : {
                "TableName": self._table_name,
                "AttributeDefinitions": attribute_definitions,
                "KeySchema": key_schema,
                "ProvisionedThroughput": {
                    'ReadCapacityUnits': self._read_units,
                    'WriteCapacityUnits': self._write_units
                }
            }
        }

    def resource_exists(self):
        client = self._context.get_client("dynamodb")
        try:
            desc = client.describe_table(TableName=self._table_name)
            return True
        except Exception as e:
            print("Table %s does not exist: %s" % (self._table_name, e))
            return False



import hashlib
import random
import json
import os
import os.path
import redleader.util as util

class Resource(object):
    def __init__(self, context, cf_params):
        super(Resource, self).__init__()
        self._context = context
        self._dependencies = []
        self._cf_params = cf_params
        self._user_id = None
        self._multiplicity_uid = None
        self._generated_id = None
        self._generated_resources = None

    def get_dependencies(self):
        return self._dependencies

    def is_static(self):
        """
        Static resources only generate cloud formation templates
        when they don't already exist. Example uses: S3 buckets, SQS queues
        """
        return False

    def add_dependency(self, dep):
        """
        Resource dependencies will be included in the cloud formation
        `DependsOn` attribute to ensure correct creation order.

        Dependencies must be added to the cluster or generated
        as subresources of included resources.
        """
        self._dependencies.append(dep)

    def get_id(self):
        """
        Each resource needs a reproducible UID that represents its state and multiplicty
        State: If a key parameter to a resource changes, it's a different resource
        Multiplicity: We need to be able to differentiate identical resources. e.g) t2.micro Instance #3
        Solution:
	  * Utilize _get_multiplicity to track # of identical resources produced
	  * Utilize _idempotent_params() to get a subset of a resource's
	    cloud formation template output, and hash it.
        Implications:
	  * get_id() cannot be used inside of _cloud_formation_template()
	    ** Instead, we'll output the placeholder {resource_id}
        """
        if self._user_id is not None:
            return self._user_id
        if self._generated_id is None:
                class_name = str(self.__class__.__name__).replace("Resource", "")
                param_hash = self._param_hash()
                if self._multiplicity_uid is None:
                    self._multiplicity_uid = Resource._get_multiplicity(class_name + param_hash)
                self._generated_id = "RL%sN%sP%s" % (class_name,
                                                     self._multiplicity_uid,
                                                     param_hash)
                if self._context.pretty_names():
                    h = hashlib.md5()
                    h.update(self._generated_id.encode('utf-8'))

                    ints = []
                    for x in range(2):
                        ints.append(int(h.hexdigest()[x * 8:(x+ 1) * 8], 16))
                    d = self._context.get_dict()

                    pretty_words = ""
                    for i in ints:
                        word = d[i % len(d)].lower().replace("'", "")
                        pretty_words += word[0].upper() + word[1:].lower()

                    self._generated_id = "%s%s%s" % (class_name, self._multiplicity_uid, pretty_words)

        return self._generated_id

    def _id_placeholder(self):
        """
        Placeholder for use in _cloud_formation_template()
        """
        return "{resource_id}"

    def _param_hash(self):
        key_params = self._idempotent_params()
        template = self._cloud_formation_template()
        extracted = {}
        for k in key_params:
            extracted[k] = template['Properties'][k]

        h = hashlib.md5()
        extracted_json = json.dumps(extracted, sort_keys=True)
        h.update(str(extracted_json).encode('utf-8'))
        return str(h.hexdigest()[0:10])

    @classmethod
    def _get_multiplicity(cls, uid):
        if(not hasattr(cls, "_multiplicity_count")):
            cls._multiplicity_count = {}
        if uid in cls._multiplicity_count:
            cls._multiplicity_count[uid] += 1
        else:
            cls._multiplicity_count[uid] = 1
        return cls._multiplicity_count[uid]

    @classmethod
    def reset_multiplicity(cls):
        cls._multiplicity_count = {}

    def _idempotent_params(self):
        """
        Returns the list of cloud formation parameters that must be the same
        in order for two RedLeader resources to refer to the same deployed resource.

        By default we assume that all parameters must be the same.

        Example: we might change an EC2 instance's security group, but want the
        RedLeader resource to refer to the same deployed server.
        """

        template = self._cloud_formation_template()
        return sorted(template['Properties'].keys())

    def iam_service_policies(self):
        """
        Return a list of objects usable by IAMRoleResource to generate
        an IAM role representing access to this resource and its sub resources
        """
        policies = []
        for resource in self.generate_sub_resources():
            policies += resource.iam_service_policies()
        policies.append(self._iam_service_policy())
        return policies

    def _iam_service_policy(self):
        raise NotImplementedError

    def generate_sub_resources(self):
        if self._generated_resources is None:
            self._generated_resources = self._generate_sub_resources()
        return self._generated_resources

    def _generate_sub_resources(self):
        """
        Generate any sub resources, if necessary
        """
        return []

    @staticmethod
    def cf_ref(resource):
        if resource is None:
            return {}
        return {"Ref": resource.get_id()}

    @staticmethod
    def cf_attr(resource, attr):
        return {"Fn::GetAtt": [ resource.get_id(), attr ]}

    @staticmethod
    def replaceValues(obj, replaceMap):
        if isinstance(obj, dict):
            for key in obj:
                if isinstance(obj[key], str):
                    obj[key] = util.multireplace(obj[key], replaceMap)
                else:
                    obj[key] = Resource.replaceValues(obj[key], replaceMap)
        if isinstance(obj, list):
            new = []
            for elem in obj:
                if isinstance(elem, str) and elem in replaceMap:
                    new.append(replaceMap[elem])
                else:
                    new.append(Resource.replaceValues(elem, replaceMap))
            return new
        return obj

    def cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        if(self.is_static() and self.resource_exists()):
            # Don't create templates for static resources that exist
            return None

        cf_template = self._cloud_formation_template()
        if cf_template is None:
            return None

        for param in self._cf_params:
            cf_template['Properties'][param] = self._cf_params[param]

        replaceMap = {"{resource_id}": self.get_id()}
        for param in cf_template['Properties']:
            cf_template['Properties'] = Resource.replaceValues(cf_template['Properties'], replaceMap)
        cf_template["DependsOn"] = []
        for dependency in self.get_dependencies():
            if(not dependency.is_static() or not dependency.resource_exists()):
                cf_template["DependsOn"].append(dependency.get_id())

        if self.is_static():
            # Don't delete static resources on cluster deletion
            cf_template["DeletionPolicy"] = "Retain"

        return cf_template

    def find_deployed_resources(self):
        """
        Finds already deployed resources that match this resource's configuration
        """
        raise NotImplementedError

class CustomUserResource(Resource):
    """
    CustomUserResource allows a cluster to provision and depend upon
    resources that aren't yet implemented programatically
    """
    def __init__(self, context, template):
        super(self, context)
        self._template = template

    def cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        return self._template



import boto3
from base64 import b64decode
import json
import subprocess
import botocore


class RedLeader(object):
    """
    Class that manages ECS cluster creation, deployment, and updates
    """
    def __init__(self, 
                 cluster_prefix,
                 image_name,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_profile=None,
                 aws_region="us-west-1"
    ):
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_profile = aws_profile
        self._aws_region = aws_region        
        self._cluster_prefix = cluster_prefix
        self._image_name = image_name
        self._session = None
        self._ecs_client = self._get_client('ecs')
        
        print("Red Leader Initialized")

    def _get_client(self, service):
        """
        Return a client configured with current credentials and region
        """
        if self._session is None:
            try:
                self._session = boto3.Session(profile_name=self._aws_profile,
                                              region_name=self._aws_region
                )
            except botocore.exceptions.ProfileNotFound:
                self._session = boto3.Session(region_name=self._aws_region)
        return self._session.client(service)

    def _task_defn_family(self):
        """
        Returns the task definiton family we'll create and look for in our account
        """
        return self._cluster_prefix + "-task"

    def _get_cluster(self):
        """
        Returns the cluster ARN for the cluster matching the setup cluster prefix
        """

        clusters = self._ecs_client.list_clusters()
        if len(clusters) is 0:
            raise Exception("No clusters found in current region.")

        cluster_arn = None
        for possible_cluster_arn in clusters['clusterArns']:
            if self._cluster_prefix in possible_cluster_arn:
                cluster_arn = possible_cluster_arn
        if cluster_arn is None:
            raise Exception("No cluster found containing cluster prefix '%s'" % self._cluster_prefix)
        print("Found cluster %s" % cluster_arn)        
        return cluster_arn

    def _get_latest_task_definition(self, cluster_arn):
        """
        Retrieves the task definition with the largest revision number for
        a given cluster.
        """
        cluster_desc = self._ecs_client.describe_clusters(clusters=[cluster_arn])
        task_defns = self._ecs_client.list_task_definitions(familyPrefix=self._task_defn_family())
        task_defn_arns = task_defns['taskDefinitionArns']

        print("Found %d matching task definitions" % len(task_defn_arns))

        if len(task_defn_arns) is 0:
            raise Exception("No matching task definitions found in cluster '%s'" % cluster_arn)

        max_rev = 0
        task_defn = None
        for task_defn_arn in task_defn_arns:
            task_desc = self._ecs_client.describe_task_definition(
                taskDefinition=task_defn_arn)
            if task_desc['taskDefinition']['revision'] > max_rev:
                max_rev = task_desc['taskDefinition']['revision']
                task_defn = task_desc['taskDefinition']
    
        if task_defn is None:
            raise Exception("Couldn't sort task definitions by revision")
        print("Found latest task definition with revision %d" % max_rev)
        return task_defn

    def _update_container_definition_image_tag(self, cont_defn, new_tag):
        """
        Update a container definition's image to utilize a certain tag
        """
        # Update image property
        new_cont_defn = cont_defn
        # Strip out current revision tag if present
        image_arr = new_cont_defn['image'].split(":")
        if len(image_arr) > 1:
            image_arr = image_arr[:-1]
        new_cont_defn['image'] = ":".join(image_arr + [new_tag])
        return new_cont_defn

    def _get_default_service(self):
        """
        Return the default service based on self._cluster_prefix
        """
        cluster_arn = self._get_cluster()
        services = self._ecs_client.list_services(
            cluster = cluster_arn)
        if len(services['serviceArns']) < 1:
            print(services)
            raise Exception('No services found for cluster %s' % cluster_arn)
        return services['serviceArns'][0]
        
    def _update_service_container_definition(self, service_arn, task_role_arn, cont_defn):
        """
        Update the service's container definition by generating
        a new task definition and updating the service to use that definition
        """
        cluster_arn = self._get_cluster()
        res = self._ecs_client.register_task_definition(
            family = self._task_defn_family(),
            taskRoleArn = task_role_arn,
            containerDefinitions = [cont_defn]
        )
        new_task_defn_arn = res['taskDefinition']['taskDefinitionArn']

        res = self._ecs_client.update_service(
            service = service_arn,
            cluster = cluster_arn,
            taskDefinition = new_task_defn_arn
        )
        return res
        
        
    def update_default_service_image_tag(self, tag):
        """
        Update the default service's image tag to the given tag
        """
        cluster_arn = self._get_cluster()
        task_defn = self._get_latest_task_definition(cluster_arn)
        if len(task_defn['containerDefinitions']) < 1:
            raise Exception("Too few container definitions for task")
        cont_defn = task_defn['containerDefinitions'][0]
        new_cont_defn = self._update_container_definition_image_tag(cont_defn, tag)

        service_arn = self._get_default_service()
        res = self._update_service_container_definition(
            service_arn,
            task_defn['taskRoleArn'],
            new_cont_defn)
        print("Updated default service image tag to "+tag)

        return res

    def _ecr_get_docker_login(self):
        ecr_client = self._get_client('ecr')
        result = ecr_client.get_authorization_token()
        for auth in result['authorizationData']:
            auth_token = b64decode(auth['authorizationToken']).decode()
            username, password = auth_token.split(':')
            return {"username": username,
                    "password": password,
                    "proxyEndpoint": auth['proxyEndpoint']}

    def _ecr_push_build(self, container_name, filepath):
        """
        Build & push the latest docker build with the given name to ECR.

        Tags the current build with the latest git commit
        """
        
        version = subprocess.Popen("git rev-parse --short HEAD", shell=True,
                               stdout=subprocess.PIPE).stdout.read().decode('utf-8')
        version = "v_" + version.strip()
        print("Git repo version: "+version)
        print("Building docker container %s:%s located at %s" % (container_name, version, filepath))
        res = subprocess.Popen("docker build -t %s:%s %s" % (container_name, version, filepath),
                               shell=True,
                               stdout=subprocess.PIPE).stdout.read().decode('utf-8')
        print(res)
        
        res = subprocess.Popen("docker tag %s:%s %s:%s" % 
                               (container_name, version, self._image_name, version),
                               shell=True,
                               stdout=subprocess.PIPE).stdout.read().decode('utf-8')
        print(res)
        print("docker tag %s:%s %s:%s" % 
                               (container_name, version, self._image_name, version))

        docker_auth = self._ecr_get_docker_login()
        login_str = ('docker login -u %s -p %s -e none %s' %
                     (docker_auth['username'],
                      docker_auth['password'],
                      docker_auth['proxyEndpoint']))

        res = subprocess.Popen(login_str + 
                               ("&& docker push %s:%s" % (self._image_name, version)),
                               shell=True,
                               stdout=subprocess.PIPE).stdout.read().decode('utf-8')
        print(res)
        return version



import re
import sys

def multireplace(string, replacements):
    substrs = sorted(replacements, key=len, reverse=True)
    regexp = re.compile('|'.join(map(re.escape, substrs)))
    return regexp.sub(lambda match: replacements[match.group(0)], string)

def print_progress(current, max_progress=20, progress_char="="):
    sys.stdout.write("[")
    for x in range(current % max_progress):
        sys.stdout.write("=")
    for x in range(max_progress - (current % max_progress)):
        sys.stdout.write(" ")
    sys.stdout.write("]")
    sys.stdout.write("\r")
    sys.stdout.flush()



import json
import time
import sys
import boto3
import urllib
import os.path

import botocore.exceptions

from .resources import Resource
from .exceptions import MissingDependencyError, OfflineContextError
import util

class Cluster(object):
    def __init__(self, cluster_class, context, pretty_names=True):
        self._cluster_class = cluster_class
        self._context = context
        self._resources = {}
        self._pretty_names = pretty_names

    def add_resource(self, resource):
        for sub_resource in resource.generate_sub_resources():
            self.add_resource(sub_resource)
        self._resources[resource.get_id()] = resource

    def validate(self):
        for resource_id in self._resources:
            resource = self._resources[resource_id]
            for dependency in resource.get_dependencies():
                x = dependency.get_id()
                if x not in self._resources:
                    print("Missing dependency")
                    print(resource.get_id())
                    print(x)
                    print(self._resources.keys())
                    raise MissingDependencyError(
                        source_resource= resource.get_id(),
                        missing_resource= dependency.get_id()
                    )

    def _mod_identifier(self, ident):
        """
        Add the cluster class onto the cloud formation identifier
        """
        return self._cluster_class + ident

    def _cluster_mod_identifiers(self, tmpl, replaceMap=None):
        """
        Modify all cloud formation identifiers so that they're unique to this cluster class
        """
        if replaceMap is None:
            replaceMap = {}
            for k in self._resources:
                replaceMap[k] = self._mod_identifier(k)

        if isinstance(tmpl, str):
            tmpl = util.multireplace(tmpl, replaceMap)
        elif isinstance(tmpl, dict):
            for k in tmpl:
                tmpl[k] = self._cluster_mod_identifiers(tmpl[k], replaceMap)
        elif isinstance(tmpl, list):
            for idx in range(len(tmpl)):
                tmpl[idx] = self._cluster_mod_identifiers(tmpl[idx], replaceMap)
        return tmpl

    def cloud_formation_template(self):
        self.validate()
        Resource.reset_multiplicity()
        templates = {}
        for resource_id in self._resources:
            resource = self._resources[resource_id]
            template = resource.cloud_formation_template()
            if template is not None:
                templates[self._mod_identifier(resource_id)] = \
                    self._cluster_mod_identifiers(template)
        return {
               "AWSTemplateFormatVersion": "2010-09-09",
               "Resources": templates
        }

    def estimate_template_cost(self):
        template = self.cloud_formation_template()
        client = self._context.get_client('cloudformation')
        return client.estimate_template_cost(TemplateBody=json.dumps(template))['Url']

    def deploy(self):
        client = self._context.get_client('cloudformation')
        return client.create_stack(
            StackName=self._cluster_class,
            TemplateBody=json.dumps(self.cloud_formation_template()),
            Capabilities=['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM']
        )

    def blocking_deploy(self, verbose=False):
        self.deploy()
        if verbose:
            print("Cluster creation in progress")
        i = 0
        while self.deployment_status() == "CREATE_IN_PROGRESS":
            i = (i + 1)
            if verbose:
                util.print_progress(i)
            time.sleep(5)
        if verbose:
            print("Cluster successfully created")
        return self.deployment_status()


    def deployment_status(self):
        client = self._context.get_client('cloudformation')
        response = client.describe_stacks(
                StackName=self._cluster_class
        )
        # Possible statuses:
        # 'StackStatus': 'CREATE_IN_PROGRESS'|'CREATE_FAILED'|'CREATE_COMPLETE'|'ROLLBACK_IN_PROGRESS'|'ROLLBACK_FAILED'|'ROLLBACK_COMPLETE'|'DELETE_IN_PROGRESS'|'DELETE_FAILED'|'DELETE_COMPLETE'|'UPDATE_IN_PROGRESS'|'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'|'UPDATE_COMPLETE'|'UPDATE_ROLLBACK_IN_PROGRESS'|'UPDATE_ROLLBACK_FAILED'|'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS'|'UPDATE_ROLLBACK_COMPLETE'|'REVIEW_IN_PROGRESS',
        return response['Stacks'][0]['StackStatus']

    def delete(self):
        client = self._context.get_client('cloudformation')
        return client.delete_stack(
            StackName=self._cluster_class
        )

    def blocking_delete(self, verbose=False):
        self.delete()
        try:
            i = 0
            if verbose:
                print("Cluster deletion in progress")
            while self.deployment_status() == "DELETE_IN_PROGRESS":
                i += 1
                if verbose:
                    util.print_progress(i)
                time.sleep(5)
            if verbose:
                print("Cluster successfully deleted")
            return self.deployment_status()
        except botocore.exceptions.ClientError:
            if verbose:
                print("Stack fully deleted, could not obtain deployment status")
            return None

    def cluster_exists(self):
        """
        Find resources for this cluster that have already deployed
        """
        try:
            status = self.deployment_status()
            return True
        except Exception as e:
            print("Cluster may not exist. Encountered error %s" % e)
            return False

    def cloud_formation_deploy(self):
        """
        TODO
        """
        raise NotImplementedError

class Context(object):
    def __init__(self, **kwargs):
        self._dict = None
        return

    def get_dict(self):
        """ Returns an english dictionary. Useful for pretty hashing"""
        if self._dict is not None:
            return self._dict
        dict_path = "/tmp/redleader_dict.txt"
        if(os.path.isfile(dict_path)):
            with open(dict_path, 'r') as f:
                self._dict = f.read().split("\n")
        else:
            print("Downloading fresh redleader dictionary")
            url = "https://svnweb.freebsd.org/csrg/share/dict/words?view=co&content-type=text%2Fplain&revision=61569"
            openfun = None
            if hasattr(urllib, "request"):
                # Python 3.x
                openfun = urllib.request.urlopen
            else:
                # Python 2.x
                openfun = urllib.urlopen

            response = openfun(url)
            dict_text = response.read().decode('utf-8')
            with open(dict_path, 'w') as f:
                f.write(dict_text)
            self._dict = dict_text.split("\n")
        return self._dict

class OfflineContext(Context):
    def __init__(self, **kwargs):
        super().__init__()

    def pretty_names(self):
        return False

    def get_session(self):
        raise OfflineContextError(action="get_session")

    def get_account_id(self):
        return "offline_context_account_id"

    def get_region(self):
        return "us-west-1"

    def get_client(self, service):
        raise OfflineContextError(action="get_client")

class AWSContext(Context):
    """
    Context manager for RedLeader, managing AWS sessions and clients.
    """

    def __init__(self,
                 aws_profile=None,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_region="us-west-1",
                 pretty_names=True
    ):
        super(AWSContext, self).__init__()
        self._aws_profile = aws_profile
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_region = aws_region
        self._pretty_names = pretty_names
        self._clients = {}
        self._account_id = None

        try:
            self._session = boto3.Session(profile_name=self._aws_profile,
                                          region_name=self._aws_region)
        except botocore.exceptions.NoCredentialsError:
            self._session = boto3.Session(region_name=self._aws_region)

    def get_session(self):
        return self._session

    def get_region(self):
        return self._aws_region

    def get_account_id(self):
        if self._account_id is None:
            self._account_id = self.get_client('sts').get_caller_identity().get('Account')
        return self._account_id

    def pretty_names(self):
        return self._pretty_names

    def get_client(self, client_type):
        if client_type not in self._clients:
            self._clients[client_type] = self._session.client(client_type)
        return self._clients[client_type]



from redleader.resources import Resource
import redleader.exceptions as exceptions

class CloudWatchLogs(Resource):
    """
    Resource modeling a Cloud Watch Logs resource.

    Currently only useful for generating access policies.
    """
    def __init__(self,
                 context,
                 cf_params={}
    ):
        super(CloudWatchLogs, self).__init__(context, cf_params)

    def is_static(self):
        return True

    def resource_exists(self):
        return True

    def get_id(self):
        return "cloudWatchLogs"

    def _iam_service_policy(self):
        return {"name": "cloudwatchlogs",
                "params": {}
        }

    def _cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        return None



from redleader.resources import Resource
import redleader.exceptions as exceptions

class S3BucketResource(Resource):
    """
    Resource modeling an S3 Bucket
    """
    def __init__(self,
                 context,
                 bucket_name,
                 cf_params={}
    ):
        super().__init__(context, cf_params)
        self._bucket_name = self._sanitize(bucket_name)

    def _sanitize(self, name):
        return name.lower().replace("-", "").replace("_", "")

    def is_static(self):
        return True

    def get_id(self):
        return "s3Bucket%s" % self._bucket_name

    def _iam_service_policy(self):
        return {"name": "s3",
                "params": {"bucket_name": self._bucket_name}}

    def _cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        return {
            "Type" : "AWS::S3::Bucket",
            "Properties" : {
                "BucketName" : self._bucket_name,
                "Tags": [{"Key": "redleader-resource-id", "Value": self.get_id()}]
            }
        }

    def resource_exists(self):
        try:
            client = self._context.get_client('s3')
        except exceptions.OfflineContextError:
            return False
        try:
            client.head_bucket(Bucket=self._bucket_name)
            return True
        except Exception as e:
            return False



import os.path
import os
import time
import tarfile
import random
import redleader.util as util

class CodeDeployManager(object):
    def __init__(self, context):
        self._context = context

    def create_code_deploy_package(self, path):
        n = "code_deploy_package%s.tgz" % random.randrange(0, 10000000)
        self.make_tarfile("./%s" % n, path)
        return n

    @staticmethod
    def make_tarfile(output_filename, source_dir):
        with tarfile.open(output_filename, "w:gz") as tar:
            files = os.listdir(source_dir)
            for f in files:
                tar.add(os.path.join(source_dir, f), arcname=f)

    def upload_package(self, bucket_name, path, name):
        print("Uploading %s to bucket %s/%s" % (path, bucket_name, name))
        client = self._context.get_client('s3')
        f = client.upload_file(path, bucket_name, name)
        # TODO. Configure bucket permissions so that an
        #    IAM policy with s3::GetObject is sufficient.
        #    Then we can remove setting ACL to authenticated-read
        client.put_object_acl(Bucket=bucket_name, Key=name,
                              ACL='authenticated-read')
        return f

    def create_deployment(self, application_name, deployment_group_name,
                          path, bucket_name, version="0", blocking=True, verbose=True):
        package = self.create_code_deploy_package(path)
        x = self.upload_package(bucket_name, "./%s" % package, package)
        client = self._context.get_client('codedeploy')
        res = client.create_deployment(
            applicationName=application_name,
            deploymentGroupName=deployment_group_name,
            revision={
                'revisionType': 'S3',
                's3Location': {
                    'bucket': bucket_name,
                    'key': package,
                    'bundleType': 'tgz',
                }
            })
        os.remove(package)
        deploymentId = res['deploymentId']
        if not blocking:
            return deploymentId

        i = 0
        if verbose:
            print("Deploying..")
        while client.get_deployment(deploymentId=deploymentId)['deploymentInfo']['status'] in ["Created", "InProgress"]:
            i += 1
            if verbose:
                util.print_progress(i)
            time.sleep(2)
        if verbose:
            print("Deployment Status: %s" % client.get_deployment(deploymentId=deploymentId)['deploymentInfo']['status'])
        return (deploymentId, client.get_deployment(deploymentId=deploymentId)['deploymentInfo']['status'])



from redleader.resources import Resource
import random
import os.path
import pkg_resources
import json

class IAMInstanceProfileResource(Resource):
    """
    Create an IAM Instance Profile with access to the given resources.
    Automatically creates an IAMRoleResource if provided with resource list

    A list of ResourcePermissions can be provided, in which case an IAMRoleResource
    will be automatically generated. Otherwise, an IAMRoleResource can be provided directly.
    """

    def __init__(self, context, permissions=[], roles=[], services=['ec2.amazonaws.com']):
        super(IAMInstanceProfileResource, self).__init__(context, {})
        self._permissions = permissions
        self._generated_role = None
        self._roles = roles
        self._services = services

        for permission in self._permissions:
            self.add_dependency(permission.resource)

        if len(self._roles) is not 0:
            for role in self._roles:
                self.add_dependency(role)

    def _generate_sub_resources(self):
        self._generated_role = IAMRoleResource(self._context,
                                               self._permissions,
                                               services=self._services,
                                               policies=[]
        )
        self._roles.append(self._generated_role)
        self.add_dependency(self._generated_role)
        return [self._generated_role]

    def _cloud_formation_template(self):
        return {
            "Type": "AWS::IAM::InstanceProfile",
            "Properties": {
                "Path": "/redleader/",
                "Roles": [Resource.cf_ref(x) for x in self._roles]
            }
        }

class IAMRoleResource(Resource):
    """
    Create an IAM role with access to the given resources.
    `policies` should be a list of IAMPolicyResource objects
    """
    def __init__(self, context, permissions, services=[], policies=[], policy_arns=[]):
        super(IAMRoleResource, self).__init__(context, {})
        self._permissions = permissions
        self._services = services
        self._policies = policies
        self._generated_policy = None
        self._policy_arns = policy_arns

        for permission in self._permissions:
            self.add_dependency(permission.resource)

        for policy in self._policies:
            self.add_dependency(policy)

    @staticmethod
    def assumeRolePolicyDocument(services):
        return {
            "Version" : "2012-10-17",
            "Statement": [
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {
                        "Service": services
                    },
                    "Action": "sts:AssumeRole"
                }
            ]
        }

    def _cloud_formation_template(self):
        return {
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument":
                   IAMRoleResource.assumeRolePolicyDocument(self._services),
                "ManagedPolicyArns": [Resource.cf_ref(x) for x in self._policies] +\
                  self._policy_arns,
                "Path": "/redleader/",
                "RoleName": self._id_placeholder()
            }
        }
        return "IAM Role template"

    def _generate_sub_resources(self):
        """
        Generate a single sub-policy representing access to provided resources
        """
        if len(self._permissions) == 0:
            return []

        self._generated_policy = IAMPolicyResource(self._context, self._permissions)
        self.add_dependency(self._generated_policy)
        self._policies.append(self._generated_policy)
        return [self._generated_policy]

class IAMPolicyResource(Resource):
    def __init__(self, context, permissions):
        super(IAMPolicyResource, self).__init__(context, {})
        self._permissions = permissions

    def _cloud_formation_template(self):
        return {
            "Type" : "AWS::IAM::ManagedPolicy",
            "Properties" : {
                "Description" : self._id_placeholder(),
                "Path": "/redleader/",
                "PolicyDocument": generate_policy_document(
                    self._context, self._permissions)
            }
        }
        return "IAM Policy template"

def generate_policy_document(context, permissions):
    policy_statements = []
    for permission in permissions:
        resource = permission.resource
        # TODO: Differentiate between read, write, readwrite
        for service in resource.iam_service_policies():
            policy = create_iam_policy(context, service)
            policy_statements += policy
    return {
        "Version" : "2012-10-17",
        "Statement": policy_statements
    }

def load_policy_file(name):
    resource_package = __name__
    resource_path = '/'.join(('policies', '%s.policy' % name))
    return pkg_resources.resource_string(resource_package, resource_path).decode('utf-8')

def create_iam_policy(context, service):
    template = load_policy_file(service['name'])
    policy = template

    for param in service['params']:
        policy = policy.replace("{%s}" % param, service['params'][param])

    builtin_params = {
        "account_id": context.get_account_id(),
        "aws_region": context.get_region()
    }
    for param in builtin_params:
        policy = policy.replace("{%s}" % param, builtin_params[param])

    return json.loads(policy)



from .resource import *
from .dynamodb import *
from .s3 import *
from .sqs import *
from .ec2 import *
from .cloudwatch import *
from .codedeploy import *
from .iam import *
from .permissions import *



class ResourcePermission(object):
    def __init__(self, resource):
        self.resource=resource

class WritePermission(ResourcePermission):
    def __init__(self, resource):
        super(WritePermission, self).__init__(resource)

class ReadPermission(ResourcePermission):
    def __init__(self, resource):
        super(ReadPermission, self).__init__(resource)

class ReadWritePermission(ResourcePermission):
    def __init__(self, resource):
        super(ReadWritePermission, self).__init__(resource)



from setuptools import setup, find_packages

install_requires = [
    'botocore',
      'boto3==1.4.2'
]

setup(
    name='RedLeader',
    version='0.1.5',
    description='An automated EC2 & ECS deployment platform',
    author='Morgan McDermott',
    author_email='morganmcdermott@gmail.com',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
        ]
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)



from redleader.resources import Resource
import redleader.exceptions as exceptions

class SQSQueueResource(Resource):
    """
    Resource modeling an S3 Bucket
    """
    def __init__(self,
                 context,
                 queue_name,
                 cf_params={}
    ):
        super(SQSQueueResource, self).__init__(context, cf_params)
        self._queue_name = queue_name

    def is_static(self):
        return True

    def get_id(self):
        return "s3Queue%s" % self._queue_name.replace("-", "").replace("_", "")

    def _iam_service_policy(self):
        return {"name": "sqs",
                "params": {"queue_name": self._queue_name}}


    def _cloud_formation_template(self):
        """
        Get the cloud formation template for this resource
        """
        return {
            "Type" : "AWS::SQS::Queue",
            "Properties" : {
                #"DelaySeconds": Integer,
                #"MaximumMessageSize": Integer,
                #"MessageRetentionPeriod": Integer,
                 "QueueName": self._queue_name
                #"ReceiveMessageWaitTimeSeconds": Integer,
                # "RedrivePolicy": RedrivePolicy,
                # "VisibilityTimeout": Integer
              }
        }

    def resource_exists(self):
        try:
            client = self._context.get_client('s3')
        except exceptions.OfflineContextError:
            return False

        client = self._context.get_client('sqs')
        queues = client.list_queues(QueueNamePrefix=self._queue_name)
        if 'QueueUrls' in queues and len(queues['QueueUrls']) > 0:
            return True
        return False
